#include <emma/lib/util.h>
#pragma hdrstop
#include "rexm.h"
#include "memcode.h"

#define TEST 0
#define MAX_INT 0x7FffFFff

static char		EscapeChar='/';
static char		ch1, ch2, ch3;
static const char *	src=NULL;
static int		CurveCount=0;
static s_codegen	Code;

static void		foresee();
static void		advance();
static void		exit_error(char*msg);
static void		comp_charset(boolean isShell, char set[]);
static boolean		read_int(int*Ret);
static void		comp_seq();
static void		comp_node();

/*=====================
 Public
=====================*/
boolean PAT_Compile(const char*Image, s_bytemem& Pattern)
{
	src=Image;
	CurveCount=0;
	foresee();

	Code.Reset();
	t_label node1=Code.Label();
	comp_seq();
	Code.PutChar(OP_RET);
	Code.PutChar(OP_NULL);
	Code.GotoLabel(node1);
	Code.PutInt(CurveCount);
	Pattern=Code.GetCode();
	return true;
}

/*=====================
 Expression parser
=====================*/
static void foresee()
{
	ch1=src[0];
	ch2=ch1 ? src[1]: NUL;
	ch3=ch2 ? src[2]: NUL;
}

static void advance()
{
	if(ch1)
	{
		src++;
		foresee();
	}
}

static void exit_error(char*msg)
{
	printf("Error: %s\n", msg);
	exit(0);
}

static boolean read_int(int*Ret)
{
	vstring buf;
	while(is_decimal(ch1))
	{
		buf.putch(ch1);
		advance();
	}
	if(buf.is_empty())
		return false;

	*Ret=atoi(buf);
	return true;
}

static boolean is_range(int c)
{
	return c=='*' || c=='+' || c=='?' || c=='{';
}

static boolean is_meta(int c)
{
	return c=='|' || c=='(' || c==')' || c=='.' || c== '^' || c== '[';
}

static void set_bit(char set[], byte bit, boolean state)
{
	byte index = bit>>3;
	byte mask = 1 << (bit&0x7);

	if(state)
		set[index] |= mask;
	else
		set[index] &= ~mask;
}

/*=====================
 Regular expression compiler
=====================*/
static void comp_seq()
{
	TList<t_label> Switch;

	while(1)
	{
		t_label seq=Code.Label();

		Switch.Push( seq );

		while(ch1 && ch1!='|' && ch1!=')')
			comp_node();

		Code.Freeze( seq );

		if(ch1!='|') break;
		advance();
	}

	if(Switch > 1)
	{
		t_label nextnode=Code.Label();

		Code.GotoLabel( Switch[0] );
		Code.PutChar(OP_ALT);

		for(int i=0; i<Switch; i++)
		{
			Code.PutRef(Switch[i]);
		}

		Code.PutInt(BADI);	// terminate the list

			// Insert Jump operator into upper blocks
		for(int i=1; i<Switch; i++)
		{
			Code.GotoLabel(Switch[i]);
			Code.PutChar(OP_JUMP);
			Code.PutRef(nextnode);
		}

		Code.GotoEnd();
		Code.SetLabel(nextnode);
	}
}

static void comp_charset(boolean isShell, char set[])
{
	byte initbyte=0;
	boolean bitstate=true;

	if(ch1=='^')
	{
		advance();
		bitstate=false;
		initbyte=0xFF;
	}

	if(isShell) memset(set, initbyte, 256/8);

	while(ch1 && ch1!=']')
	{
		if(ch1==EscapeChar)
		{
			advance();
			set_bit(set, SAP_EscapeChar(ch1), bitstate);
			advance();
		} else if(ch2=='-' && ch3)
		{
			for(int i=ch1; i<ch3; i++)
				set_bit(set, i, bitstate);
			advance();
			advance();
		} else if(ch1=='[')
		{
			advance();
			comp_charset(false, set);
			advance();
		} else
		{
			set_bit(set, ch1, bitstate);
			advance();
		}
	}

	if(!ch1)exit_error("expected ]");
}

static void comp_node()
{
	t_label prevnode=Code.Label();
	t_label nextnode=Code.Label();

	switch(ch1)
	{
	case '(':{
		int iCurve= ++CurveCount;
		Code.PutChar(OP_GROUPL);
		Code.PutInt(iCurve);
		advance();
		comp_seq();
		if(ch1!=')')exit_error("syntax error");
		Code.PutChar(OP_GROUPR);
		Code.PutInt(iCurve);
		advance();
		break;}

	case '.':
		advance();
		Code.PutChar(OP_NOTEOL);
		break;

	case '^':
		advance();
		Code.PutChar(OP_EDGEOFLINE);
		break;

	case '[':{
		char set[32];
		advance();
		comp_charset(true, set);
		advance();
		Code.PutChar(OP_CHARSET);
		Code.Put(set, 32);
		break;}

	default: {
		if(ch1 && is_range(ch2))
		{
			Code.PutChar(OP_PATTERN);
			Code.PutInt(1);
			Code.PutChar(ch1);
			advance();
			break;
		}

		vstring match;
		while(ch1 && !is_meta(ch1) && !is_range(ch2))
		{
			if(ch1==EscapeChar)
			{
				advance();
				match.putch(SAP_EscapeChar(ch1));
				advance();
				continue;
			}
			match.putch(ch1);
			advance();
		}

		Code.PutChar(OP_PATTERN);
		Code.PutInt(match.length());
		Code.Put(match, match.length());
		break;}
	}

	if(is_range(ch1))
	{		// Consider this node as a stand alone pattern
			// and so set its code terminator
		Code.PutChar(OP_RET);
		Code.Freeze(prevnode);

		int minnum=1;
		int maxnum=1;

		switch(ch1)
		{
		case'*':
			minnum=0;
			maxnum=MAX_INT;
			break;

		case'+':
			minnum=1;
			maxnum=MAX_INT;
			break;

		case'?':
			minnum=0;
			maxnum=1;
			break;

		case'{':
			advance();
			if(!read_int(&minnum))
				minnum=1;

			if(ch1==',')
			{
				advance();
				if(!read_int(&maxnum))
					maxnum=MAX_INT;
			} else	
			{
				maxnum=minnum;
			}

			if(ch1!='}')
				exit_error("} expected");
			break;
		}

		boolean IsGreedy=true;
		advance();
		if(ch1=='?')
		{
			IsGreedy=false;
			advance();
		}

			// Insert the range operator just before last pattern
		Code.GotoLabel(prevnode);
		Code.PutChar(OP_RANGE);
		Code.PutInt(minnum);
		Code.PutInt(maxnum);
		Code.PutChar(IsGreedy);
		Code.PutRef(nextnode);
	}

	Code.GotoEnd();
	Code.SetLabel(nextnode);
}

#if TEST
/*=====================
 Testing procedure
=====================*/
void main()
{
	struct {char*text; char*re; char*params;} testtab[]=
	{
	{"cock", "[A-Z]*", ""},
	{"cock", "cock|pussy", "cock"},
	{"****PopZX\nTomb Raider4*****", "(Push|Pop|(Romeo-[XY]{30,100}))[A-Z]+^(Tomb Raider)*?[a-z0-9]+", "PopZX\nTomb Raider4"},
	{"ccc", "&", NULL},
	{"___ccccc___", "c+", "ccccc"},
	{"___C___", "c+", NULL},
	{"Alpha", "[a-z]+", "lpha"},
	{"32043$&edowp", "&", "&"},
	{"rrrrrrr4839banana3()\nttttt", "([0-9][a-z0-9[^12]]+)(/(/)^)", "4839banana3()\n"},
	{"4839ALPHA()", "[0-9A-Z[^12]]+", "4839ALPHA"},
	{"4839ALPHA()", "[0-9A-Z[^19]]+", "483"},
	};

	for(int k=0; k<LEN(testtab); k++)
	{
		s_bytemem pat;
	
		if(!PAT_Compile(testtab[k].re, pat))
		{
			printf("Cannot compile: <%s>\n", testtab[k].re);
			return;
		}
	
		printf("SEARCH: <%s> in <%s>\n", testtab[k].re, testtab[k].text);

		TList<ParamRec> Params;
		if(!PAT_Search(pat, testtab[k].text, Params))
		{
			if(testtab[k].params==NULL)
			{
				printf("\tOk (not found)\n");
				continue;
			}

			printf("*This pattern must match\n");
			printf("%s\n", String(PAT_Describe(pat)));
			return;
		}

		if(testtab[k].params==NULL)
		{
			printf("*This pattern must not match to <%s>\n", String(PAT_Param(Params[0])));
			printf("%s\n", String(PAT_Describe(pat)));
			return;
		}

		const char*scan=testtab[k].params;
		char param[256];
		for(int j=0; SAP_ScanForToken(&scan, ";", param); j++)
		{
			vstring buf=PAT_Param(Params[j]);

			if(buf==param) continue;

			printf("param[%i] <%s> is not equal to <%s>\n", j, param, String(buf));
			printf("%s\n", String(PAT_Describe(pat)));
			return;
		}

		printf("\tOk found <%s>\n", String(PAT_Param(Params[0])));
	}
	printf("Test status: done\n");
}
#endif
